// LCD-Routines for Graphical-Displays with Samsung S6B0724 controller
// Johann Gysin
//
// History
// 22.10.2003  Johann Gysin  First Version
// 11.11.2003  Johann Gysin  Version without framebuffer (1k less RAM but slower)
// 28.09.2004  Johann Gysin  Very small modifications for the eZ80 (include eZ80.h and modified portnames...)

#include "glcd.h"
#include "font.h"
#include "stringhdl.h"
#include "hardware.h"
#include "picture.h"
#include <eZ80.h>

// Choose how the display is connected
#define PDataDDR    PA_DDR // Display Data; Port Direction Register (Input/Output)
#define PDataOutReg PA_DR  // Display Data; Port Output Register
#define PDataInReg  PA_DR  // Display Data; Port Input Register
#define PSDDR       PC_DDR // Display Setup; Port Direction Register (Input/Output)
#define PSOutReg    PC_DR  // Display Setup; Port Output Register
#define PSInpReg    PC_DR  // Display Setup; Port Input Register
#define PDI         B0     // Bit for D/I (1=Data, 0=Instruction)
#define PRW         B1     // Bit for R/W (1=Read, 0=Write)
#define PE1         B2     // Bit for E1 (Enable)
#define PRES        B3     // Bit for Res (0=Display Reset)

// Display width & height in pixels
#define Width       128
#define Height      64

void GLCD_WaitTime(int Time);
void GLCD_Wait(unsigned char controller);
void GLCD_Init(void);
void GLCD_Off(void);
void GLCD_SetStartLine(unsigned char controller, unsigned char line);
void GLCD_ReverseDisplay(unsigned char controller, unsigned char yesorno);
void GLCD_SetContrast(unsigned char controller, unsigned char contrast);
void GLCD_SetPage(unsigned char controller, unsigned char page);
void GLCD_SetYAddress(unsigned char controller, unsigned char YAddress);
void GLCD_WriteDisplayData(unsigned char controller, unsigned char data);
void GLCD_WriteDisplayByte(unsigned char line, unsigned char page, unsigned char databyte);
unsigned char GLCD_ReadDisplayByte(unsigned char page, unsigned char line);
void GLCD_Clear(void);
void GLCD_SetPixel(unsigned char x, unsigned char y);
void GLCD_ResetPixel(unsigned char x, unsigned char y);
void GLCD_CalculatePageBitShift(unsigned char y, unsigned char* page, unsigned char * BitShift);
unsigned char GLCD_WriteChar(unsigned char page, unsigned char line, unsigned char character);
unsigned char GLCD_OverWriteChar(unsigned char page, unsigned char line, unsigned char character);
void GLCD_WriteStr(unsigned char page, unsigned char line, char* str);
void GLCD_OverWriteStr(unsigned char page, unsigned char line, char* str);
void GLCD_WriteInt(unsigned char page, unsigned char line, int number);
void GLCD_OverWriteInt(unsigned char page, unsigned char line, int number);
void GLCD_DrawLine(int x1, int y1, int x2, int y2);
void GLCD_DrawBox(unsigned char StartX, unsigned char StartY, unsigned char EndX, unsigned char EndY);
void GLCD_FillBox(unsigned char StartX, unsigned char StartY, unsigned char EndX, unsigned char EndY);
void GLCD_ClearBox(unsigned char StartX, unsigned char StartY, unsigned char EndX, unsigned char EndY);
void GLCD_ShowPicture(unsigned char index);
void GLCD_circlePoints(unsigned char cx, unsigned char cy, unsigned char x, unsigned char y);
void GLCD_DrawCircle(unsigned char xCenter, unsigned char yCenter, unsigned char radius);
void GLCD_WriteCommand(unsigned char controller, unsigned char commandbyte);
void GLCD_BoundaryFill(int x, int y);
void GLCD_FloorFill8(int x, int y);
void GLCD_ClearLine(unsigned char line);
void GLCD_InvertPage(unsigned char line, unsigned char xStart, unsigned char xEnd);

void GLCD_WaitTime(int Time)
  {
  volatile int a;
  volatile int b;

  for (a=0; a<Time; a++)
    {
    for (b=0; b<500; b++) // wait some time
      {
      }
    }

  }
  
void GLCD_Wait(unsigned char controller)
  {    
  PDataDDR = 255; // Data Port as Input
  
  do
    {
    PSOutReg    = PRES + PRW;
    PSOutReg    = PRES + PRW + controller;
    }
  while (PDataInReg & B7);
    
  PSOutReg = PRES;
  PDataDDR = 0; // Data Port as Output
  }

void GLCD_Init(void)
  {
  // Setup Ports
  PDataOutReg = 0;
  PDataDDR    = 0;                           // Port as Output
  PSOutReg    = 0   ;                        // Reset
  PSDDR       = ~(PDI+PRW+PE1+PRES);
    
  PSOutReg    = PRES;                        // No Reset
    
  GLCD_SetStartLine(PE1, 0);
  GLCD_WriteCommand(PE1, B0+B5+B7);          // (ADC) Segment Driver Direction Select  
  GLCD_WriteCommand(PE1, B6+B7);             // (SHL) Com Direction
  GLCD_WriteCommand(PE1, B0+B1+B5+B7);       // BIAS Select
  GLCD_WriteCommand(PE1, B2+B3+B5);          // Power ON (VC)
  GLCD_WaitTime(2);                          // Wait (50% rising of Vout)
  GLCD_WriteCommand(PE1, B1+B2+B3+B5);       // Power ON (VC, VR)
  GLCD_WaitTime(2);                          // Wait (>=1ms)
  GLCD_WriteCommand(PE1, B0+B1+B2+B3+B5);    // Power ON (VC, VR, VF)    
  GLCD_WriteCommand(PE1, B2+B5);             // Regulator Resistor Select  
  GLCD_WriteCommand(PE1, B0+B7);             // Regulator Reference Select
  GLCD_WriteCommand(PE1, B5);                // Regulator Reference Select
  GLCD_Clear();                              // Clear Display
  GLCD_WriteCommand(PE1, B0+B1+B2+B3+B5+B7); // Display ON
  }

void GLCD_Off(void)
  {
  GLCD_WriteCommand(PE1, B1+B2+B3+B5+B7); // Display OFF
  }

void GLCD_WriteCommand(unsigned char controller, unsigned char commandbyte)
  {
  PDataOutReg = commandbyte;
  PSOutReg    = PRES + controller;        // Enable
  PSOutReg    = PRES;
  GLCD_Wait(controller);
  }
  
void GLCD_SetStartLine(unsigned char controller, unsigned char line)
  {
  GLCD_WriteCommand(controller, B6 + line); // Set Display Start Line
  }
  
void GLCD_ReverseDisplay(unsigned char controller, unsigned char yesorno)
  {
  if (yesorno)
    {
    GLCD_WriteCommand(controller, B7+B5+B2+B1+B0);
    }
  else
    {
    GLCD_WriteCommand(controller, B7+B5+B2+B1);
    }
  }

void GLCD_SetContrast(unsigned char controller, unsigned char contrast)
  {
  GLCD_WriteCommand(controller, B7+B0);
  GLCD_WriteCommand(controller, contrast);
  }
  
void GLCD_SetPage(unsigned char controller, unsigned char page)
  {
  GLCD_WriteCommand(controller, B4+B5+B7 + page); // Set Page
  }
  
void GLCD_SetYAddress(unsigned char controller, unsigned char YAddress)
  {
  YAddress = YAddress+4; // Because of flipped image mode

  GLCD_WriteCommand(controller, B4 + (YAddress>>4));     // Y-Address 0 High Nibble
  GLCD_WriteCommand(controller, YAddress&(B0+B1+B2+B3)); // Y-Address 0 Low Nibble
  }
  
void GLCD_WriteDisplayData(unsigned char controller, unsigned char data)
  {
  PSOutReg    = PRES;
  PDataOutReg = data;
  PSOutReg    = PRES + PDI + controller;
  PSOutReg    = PRES + PDI;
  PSOutReg    = PRES;
  GLCD_Wait(controller);
  }
  
unsigned char GLCD_ReadDisplayByte(unsigned char page, unsigned char line)
  {
  unsigned char databyte;
    
  GLCD_SetPage(PE1, page);
  GLCD_SetYAddress(PE1, line);
    
  PDataDDR = 255;                    // Data Port as Input
  PSOutReg = PRES + PRW + PDI;
  PSOutReg = PRES + PRW + PDI + PE1; // Dummy Read
  PSOutReg = PRES + PRW + PDI;
  PSOutReg = PRES + PRW + PDI + PE1; // Read
  databyte = PDataInReg;
  PSOutReg = PRES;
  PDataDDR = 0;                      // Data Port as Output
    
  return(databyte);
  }
  
void GLCD_WriteDisplayByte(unsigned char page, unsigned char line, unsigned char databyte)
  {
  GLCD_SetPage(PE1, page);
  GLCD_SetYAddress(PE1, line);
  GLCD_WriteDisplayData(PE1, databyte);
  }
  
void GLCD_Clear(void)
  {
  unsigned char line;
  unsigned char page;
    
  for (line=0; line<Width; line++)
    {
    for (page=0; page<8; page++)
      {
      GLCD_WriteDisplayByte(page, line, 0); // Clear all bits on whole display
      }
    }
  }

void GLCD_InvertPicture(void)
  {
  unsigned char line;
  unsigned char page;
  unsigned char databyte;
    
  for (line=0; line<Width; line++)
    {
    for (page=0; page<8; page++)
      {
      databyte = GLCD_ReadDisplayByte(page, line);
      GLCD_WriteDisplayByte(page, line, ~databyte);
      }
    }
  }

  
void GLCD_SetPixel(unsigned char x, unsigned char y)
  {
  unsigned char page = 0;
  unsigned char BitShift = 0;
  unsigned char OldByte;
  unsigned char NewByte;
  
  if ((x<Width) && (y<Height))
    {    
    GLCD_CalculatePageBitShift(y, &page, &BitShift);
    OldByte = GLCD_ReadDisplayByte(page, x);
    NewByte = OldByte | (1 << BitShift);
    GLCD_WriteDisplayByte(page, x, NewByte);
    }
  }

char GLCD_GetPixel(unsigned char x, unsigned char y)
  {
  unsigned char page = 0;
  unsigned char BitShift = 0;
  unsigned char DisplayByte;
  
  if ((x<Width) && (y<Height))
    {    
    GLCD_CalculatePageBitShift(y, &page, &BitShift);
    DisplayByte = GLCD_ReadDisplayByte(page, x);
    if (DisplayByte & (1 << BitShift))
      {
      return(1);
      }
    else
      {
      return(0);
      }
    }
  return(0);
  }


void GLCD_ResetPixel(unsigned char x, unsigned char y)
  {
  unsigned char page = 0;
  unsigned char BitShift = 0;
  unsigned char OldByte;
  unsigned char NewByte;
    
  if ((x<Width) && (y<Height))
    {
    GLCD_CalculatePageBitShift(y, &page, &BitShift);
    OldByte = GLCD_ReadDisplayByte(page, x);
    NewByte = OldByte & ~(1 << BitShift);
    GLCD_WriteDisplayByte(page, x, NewByte);
    }
  }


void GLCD_CalculatePageBitShift(unsigned char y, unsigned char* page, unsigned char* BitShift)
  {
  *page = y >> 3;
  *BitShift = y % 8;
  }
  
unsigned char GLCD_WriteChar(unsigned char page, unsigned char line, unsigned char character)
  {
  unsigned char index;
    
  for (index=0; index<5; index++)
    {
    GLCD_WriteDisplayByte(page, line+index, font[character][index]<<1);
    }
  GLCD_WriteDisplayByte(page, line+5, 0);
    
  return(line+6);
  }

  
unsigned char GLCD_OverWriteChar(unsigned char page, unsigned char line, unsigned char character)
  {
  unsigned char databyte;
  unsigned char index;
    
  for (index=0; index<5; index++)
    {
    databyte = GLCD_ReadDisplayByte(page, line+index);
    databyte = databyte | (font[character][index]<<1);
    GLCD_WriteDisplayByte(page, line+index, databyte);
    }
  GLCD_ClearBox(line, 8<<page, line+6, 8<<page);
    
  return(line+6);
  }

void GLCD_WriteStr(unsigned char page, unsigned char line, char* str)
  {
  while (*str != '\0')
    {
    line = GLCD_WriteChar(page, line, *str);
    str++;              
    }  
  }


void GLCD_OverWriteStr(unsigned char page, unsigned char line, char* str)
  {
  while (*str != '\0')
    {
    line = GLCD_OverWriteChar(page, line, *str);
    str++;              
    }  
  }

void GLCD_DrawLine(int x1, int y1, int x2, int y2)
  {
  int y;
  int x;
  int xi;
  int mi;
  int xf;
  int mf;
  int yi;
  int yf;
  unsigned char DoYSwap = 0;
  unsigned char DoXSwap = 0;
  int xNew;
  int yNew;
    
  if ((y1-y2)>(y2-y1))
    {
    yNew = y1;
    y1 = y2;
    y2 = yNew;
    DoXSwap = 1;
    }
    
  if ((x1-x2)>(x2-x1))
    {
    xNew = x1;
    x1 = x2;
    x2 = xNew;
    DoYSwap = 1;
    }
    
  if (DoXSwap && DoYSwap)
    {
    DoXSwap = 0;
    DoYSwap = 0;
    }
    
  if ((y2-y1) > (x2-x1))
    {
    xi = x1;                     // Linienstart (Ganzzahl)
    xf = -(y2-y1);               // - Lnge Y-Teil
    mi = (x2-x1) / (y2-y1);      // LngeX / LngeY (anzahl Schritte bei x bei einem Schritt Y)
    mf = ((x2-x1) % (y2-y1))<<1; // Nachkommastelle von LngeX / LngeY * 2
      
    for (y=y1; y<=y2; y++)       // Von y1 bis y2
      {
      if (DoXSwap)
        {
        xNew = x2-xi+x1;
        }
      else
        {
        xNew = xi;
        }
        
      if (DoYSwap)
        {
        yNew = y2-y+y1;
        }
      else
        {
        yNew = y;
        }
        
      GLCD_SetPixel(xNew, yNew);
      xi = xi + mi;
      xf = xf + mf;
      if (xf > 0)                // bertrag?
        {
        xi = xi + 1;
        xf = xf - ((y2-y1)<<1);
        }
      }
    }
  else
    {
    yi = y1;                     // Linienstart (Ganzzahl)
    yf = -(x2-x1);               // - Lnge Y-Teil
    mi = (y2-y1) / (x2-x1);      // LngeX / LngeY (anzahl Schritte bei x bei einem Schritt Y)
    mf = ((y2-y1) % (x2-x1))<<1; // Nachkommastelle von LngeX / LngeY * 2
      
    for (x=x1; x<=x2; x++)       // Von x1 bis x2
      {
      if (DoYSwap)
        {
        yNew = y2-yi+y1;
        }
      else
        {
        yNew = yi;
        }
        
      if (DoXSwap)
        {
        xNew = x2-x+x1;
        }
      else
        {
        xNew = x;
        }
        
      GLCD_SetPixel(xNew, yNew);
      yi = yi + mi;
      yf = yf + mf;
      if (yf > 0)                // bertrag?
        {
        yi = yi + 1;
        yf = yf - ((x2-x1)<<1);
        }
      }
    }
  }


void GLCD_WriteInt(unsigned char page, unsigned char line, int number)
  {
  char string[10];
    
  IntToStr(number, string);
  GLCD_WriteStr(page, line, string);
  }

void GLCD_OverWriteInt(unsigned char page, unsigned char line, int number)
  {
  char string[10];
    
  IntToStr(number, string);
  GLCD_OverWriteStr(page, line, string);
  }


void GLCD_DrawBox(unsigned char StartX, unsigned char StartY, unsigned char EndX, unsigned char EndY)
  {
  GLCD_DrawLine(StartX, StartY, EndX, StartY);
  GLCD_DrawLine(StartX, EndY, EndX, EndY);
  GLCD_DrawLine(StartX, StartY, StartX, EndY);
  GLCD_DrawLine(EndX, StartY, EndX, EndY);
  }


void GLCD_FillBox(unsigned char StartX, unsigned char StartY, unsigned char EndX, unsigned char EndY)
  {
  unsigned char x;
  unsigned char y;
    
  for (x=StartX; x<=EndX; x++)
    {
    for (y=StartY; y<=EndY; y++)
      {
      GLCD_SetPixel(x, y);
      }
    }
  }


void GLCD_ClearBox(unsigned char StartX, unsigned char StartY, unsigned char EndX, unsigned char EndY)
  {
  unsigned char x;
  unsigned char y;
    
  for (x=StartX; x<=EndX; x++)
    {
    for (y=StartY; y<=EndY; y++)
      {
      GLCD_ResetPixel(x, y);
      }
    }
  }

void GLCD_ShowPicture(unsigned char index)
  {
  unsigned char page;
  unsigned char line;
    
  for (page=0; page<8; page++)
    {
    for (line=0; line<128; line++)
      {
      switch (index)
        {
        case 0: GLCD_WriteDisplayByte(page, line, ~picture[line][page]);
                break;
        }
      }
    }
  }

void GLCD_circlePoints(unsigned char cx, unsigned char cy, unsigned char x, unsigned char y)
  {
  if (x == 0)
    {
    GLCD_SetPixel(cx, cy+y);
    GLCD_SetPixel(cx, cy-y);
    GLCD_SetPixel(cx+y, cy);
    GLCD_SetPixel(cx-y, cy);
    }
    
  if (x == y)
    {
    GLCD_SetPixel(cx+x, cy+y);
    GLCD_SetPixel(cx-x, cy+y);
    GLCD_SetPixel(cx+x, cy-y);
    GLCD_SetPixel(cx-x, cy-y);
    }
    
  if (x < y)
    {
    GLCD_SetPixel(cx+x, cy+y);
    GLCD_SetPixel(cx-x, cy+y);
    GLCD_SetPixel(cx+x, cy-y);
    GLCD_SetPixel(cx-x, cy-y);
    GLCD_SetPixel(cx+y, cy+x);
    GLCD_SetPixel(cx-y, cy+x);
    GLCD_SetPixel(cx+y, cy-x);
    GLCD_SetPixel(cx-y, cy-x);
    }
  }

void GLCD_DrawCircle(unsigned char xCenter, unsigned char yCenter, unsigned char radius)
  {
  unsigned char x = 0;
  unsigned char y;
  char p;
    
  y = radius;
  p = (5-(radius<<2))>>2;
    
  GLCD_circlePoints(xCenter, yCenter, x, y);
  while(x<y)
    {
    x++;
    if (p<0)
      {
      p = p + (x<<1) + 1;
      }
    else
      {
      y--;
      p = p + ((x-y)<<1) + 1;
      }
    GLCD_circlePoints(xCenter, yCenter, x, y);
    }
  }

void GLCD_BoundaryFill(int x, int y) // Caution: Recursive => Needs a lot of stack! 
  {
  char exit = 0;
    
  if ((x<0) || (x >= Width))
    {
    exit = 1;
    }
    
  if ((y<0) || (y >= Height))
    {
    exit = 1;
    }
  
  if (!exit)
    {
    if (GLCD_GetPixel(x, y) == 0)
      {
      GLCD_SetPixel(x, y);
      GLCD_BoundaryFill(x+1, y);
      GLCD_BoundaryFill(x, y+1);
      GLCD_BoundaryFill(x-1, y);
      GLCD_BoundaryFill(x, y-1);
      }
    }      
  }

void GLCD_FloorFill8(int x, int y) // Caution: Recursive => Needs a lot of stack! 
  {
  char exit = 0;
    
  if ((x<0) || (x >= Width))
    {
    exit = 1;
    }
    
  if ((y<0) || (y >= Height))
    {
    exit = 1;
    }
  
  if (!exit)
    {
    if (GLCD_GetPixel(x, y) == 0)
      {
      GLCD_SetPixel(x, y);
      GLCD_FloorFill8(x+1, y);
      GLCD_FloorFill8(x, y+1);
      GLCD_FloorFill8(x-1, y);
      GLCD_FloorFill8(x, y-1);
      GLCD_FloorFill8(x+1, y+1);
      GLCD_FloorFill8(x-1, y+1);
      GLCD_FloorFill8(x-1, y-1);
      GLCD_FloorFill8(x+1, y-1);
      }
    }      
  }

void GLCD_ClearLine(unsigned char line)
  {
  unsigned char index;
    
  GLCD_SetPage(PE1, line);
  GLCD_SetYAddress(PE1, 0);
    
  for (index=0; index<Width; index++)
    {
    GLCD_WriteDisplayData(PE1, 0);
    }
  }

void GLCD_InvertPage(unsigned char page, unsigned char xStart, unsigned char xEnd)
  {
  unsigned char newbyte;
  unsigned char line;
    
  for (line=xStart; line<=xEnd; line++)
    {
    newbyte = GLCD_ReadDisplayByte(page, line);
    newbyte = ~newbyte;
    GLCD_WriteDisplayByte(page, line, newbyte);
    }
  }
